import os
import warnings

import numpy as np
import pytest
from flaky import flaky
from modflow_devtools.markers import excludes_platform, requires_exe

from flopy.modflow import (
    Modflow,
    ModflowBas,
    ModflowDis,
    ModflowEvt,
    ModflowGhb,
    ModflowLak,
    ModflowLmt,
    ModflowLpf,
    ModflowNwt,
    ModflowOc,
    ModflowPcg,
    ModflowRch,
    ModflowUpw,
    ModflowWel,
)
from flopy.mt3d import (
    Mt3dAdv,
    Mt3dBtn,
    Mt3dDsp,
    Mt3dGcg,
    Mt3dLkt,
    Mt3dms,
    Mt3dPhc,
    Mt3dRct,
    Mt3dSsm,
    Mt3dTob,
)

# Test loading of MODFLOW and MT3D models that come with MT3D distribution
from flopy.utils import UcnFile


@pytest.fixture
def mt3d_test_model_path(example_data_path):
    return example_data_path / "mt3d_test"


@pytest.fixture
def mf2005mt3d_model_path(mt3d_test_model_path):
    return mt3d_test_model_path / "mf2005mt3d"


@pytest.fixture
def mf2kmt3d_model_path(mt3d_test_model_path):
    return mt3d_test_model_path / "mf2kmt3d"


@pytest.fixture
def mfnwtmt3d_model_path(mt3d_test_model_path):
    return mt3d_test_model_path / "mfnwt_mt3dusgs"


@requires_exe("mf2005", "mt3dms")
def test_mf2005_p07(function_tmpdir, mf2005mt3d_model_path):
    pth = mf2005mt3d_model_path / "P07"
    namfile = "p7mf2005.nam"
    mf = Modflow.load(namfile, model_ws=pth, verbose=True, exe_name="mf2005")
    cpth = function_tmpdir / "P07"
    mf.model_ws = cpth
    mf.write_input()

    success, buff = mf.run_model(silent=False)
    assert success, f"{mf.name} did not run"

    namfile = "p7mt.nam"
    mt = Mt3dms.load(namfile, model_ws=pth, verbose=True, exe_name="mt3dms")
    # Optional keyword line is absent in this example, ensure defaults are kept
    assert mt.btn.DRYCell is False
    assert mt.btn.Legacy99Stor is False
    assert mt.btn.MFStyleArr is False
    assert mt.btn.AltWTSorb is False

    mt.model_ws = cpth
    ftlfile = "p7.ftl"
    mt.ftlfilename = ftlfile
    mt.write_input()

    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


@requires_exe("mf2000", "mt3dms")
def test_mf2000_p07(function_tmpdir, mf2kmt3d_model_path):
    pth = mf2kmt3d_model_path / "P07"
    namfile = "p7mf2k.nam"
    mf = Modflow.load(namfile, model_ws=pth, verbose=True, exe_name="mf2000")

    cpth = function_tmpdir / "P07_2K"
    mf.model_ws = cpth

    mf.write_input()

    success, buff = mf.run_model(silent=True)
    assert success, f"{mf.name} did not run"

    namfile = "p7mt.nam"
    mt = Mt3dms.load(namfile, model_ws=pth, verbose=True, exe_name="mt3dms")
    mt.model_ws = cpth
    ftlfile = "p7.ftl"
    mt.ftlfilename = ftlfile
    mt.write_input()
    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


@requires_exe("mf2000", "mt3dms")
def test_mf2000_HSSTest(function_tmpdir, mf2kmt3d_model_path):
    pth = mf2kmt3d_model_path / "HSSTest"
    namfile = "hsstest_mf2k.nam"
    mf = Modflow.load(
        namfile, model_ws=pth, version="mf2k", verbose=True, exe_name="mf2000"
    )

    cpth = function_tmpdir / "HSSTest"
    mf.model_ws = cpth

    mf.write_input()
    success, buff = mf.run_model(silent=True)
    assert success, f"{mf.name} did not run"
    namfile = "hsstest_mt.nam"
    mt = Mt3dms.load(namfile, model_ws=pth, verbose=True, exe_name="mt3dms")

    mt.model_ws = cpth
    ftlfile = "hsstest.FTL"
    mt.ftlfilename = ftlfile
    mt.write_input()

    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


@requires_exe("mf2000", "mt3dms")
def test_mf2000_mnw(function_tmpdir, mf2kmt3d_model_path):
    # cannot run this model because it uses mnw1 and there is no load for mnw1
    # this model includes block format data in the btn file

    pth = mf2kmt3d_model_path / "mnw"
    namfile = "t5mf2k.nam"
    mf = Modflow.load(namfile, model_ws=pth, verbose=True)

    cpth = function_tmpdir / "MNW"
    mf.model_ws = cpth

    namfile = "t5mt.nam"
    mt = Mt3dms.load(namfile, model_ws=pth, verbose=True)

    mt.change_model_ws(cpth)
    mt.write_input()


@requires_exe("mf2000", "mt3dms")
def test_mf2000_MultiDiffusion(function_tmpdir, mf2kmt3d_model_path):
    pth = mf2kmt3d_model_path / "MultiDiffusion"
    namfile = "p7mf2k.nam"
    mf = Modflow.load(
        namfile, model_ws=pth, version="mf2k", verbose=True, exe_name="mf2000"
    )

    cpth = function_tmpdir / "MultiDiffusion"
    mf.model_ws = cpth

    mf.write_input()
    success, buff = mf.run_model(silent=True)
    assert success, f"{mf.name} did not run"
    namfile = "P7MT.NAM"
    mt = Mt3dms.load(namfile, model_ws=pth, verbose=True, exe_name="mt3dms")
    mt.model_ws = cpth
    ftlfile = "p7.ftl"
    mt.ftlfilename = ftlfile
    mt.write_input()
    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


@requires_exe("mf2000", "mt3dms")
def test_mf2000_reinject(function_tmpdir, mf2kmt3d_model_path):
    pth = mf2kmt3d_model_path / "reinject"
    namfile = "p3mf2k.nam"
    mf = Modflow.load(
        namfile, model_ws=pth, version="mf2k", verbose=True, exe_name="mf2000"
    )

    cpth = function_tmpdir / "reinject"
    mf.model_ws = cpth

    mf.write_input()
    success, buff = mf.run_model(silent=True)
    assert success, f"{mf.name} did not run"

    namfile = "p3mt.nam"
    mt = Mt3dms.load(namfile, model_ws=pth, verbose=True, exe_name="mt3dms")

    mt.model_ws = cpth
    ftlfile = "p3.ftl"
    mt.ftlfilename = ftlfile
    mt.write_input()
    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


@requires_exe("mf2000", "mt3dms")
def test_mf2000_SState(function_tmpdir, mf2kmt3d_model_path):
    pth = mf2kmt3d_model_path / "SState"
    namfile = "SState_mf2k.nam"
    mf = Modflow.load(
        namfile, model_ws=pth, version="mf2k", verbose=True, exe_name="mf2000"
    )

    cpth = function_tmpdir / "SState"
    mf.model_ws = cpth

    mf.write_input()
    success, buff = mf.run_model(silent=True)
    assert success, f"{mf.name} did not run"

    namfile = "SState_mt.nam"
    mt = Mt3dms.load(namfile, model_ws=pth, verbose=True, exe_name="mt3dms")

    mt.model_ws = cpth
    ftlfile = "SState.ftl"
    mt.ftlfilename = ftlfile
    mt.write_input()
    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


@requires_exe("mf2000", "mt3dms")
def test_mf2000_tob(function_tmpdir, mf2kmt3d_model_path):
    pth = mf2kmt3d_model_path / "tob"
    namfile = "p7mf2k.nam"
    mf = Modflow.load(
        namfile, model_ws=pth, version="mf2k", verbose=True, exe_name="mf2000"
    )

    cpth = function_tmpdir / "tob"
    mf.model_ws = cpth

    mf.lmt6.output_file_header = "extended"
    mf.lmt6.output_file_format = "formatted"
    mf.write_input()
    success, buff = mf.run_model(silent=True)
    assert success, f"{mf.name} did not run"

    namfile = "p7mt.nam"
    mt = Mt3dms.load(
        namfile, model_ws=pth, verbose=True, exe_name="mt3dms", forgive=True
    )
    mt.model_ws = cpth
    ftlfile = "p7.ftl"
    mt.ftlfilename = ftlfile
    mt.write_input()
    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


@requires_exe("mf2000", "mt3dms")
def test_mf2000_zeroth(function_tmpdir, mf2kmt3d_model_path):
    pth = mf2kmt3d_model_path / "zeroth"
    namfile = "z0mf2k.nam"
    mf = Modflow.load(
        namfile, model_ws=pth, version="mf2k", verbose=True, exe_name="mf2000"
    )

    cpth = function_tmpdir / "zeroth"
    mf.model_ws = cpth

    mf.write_input()
    success, buff = mf.run_model(silent=True)
    assert success, f"{mf.name} did not run"

    namfile = "z0mt.nam"
    mt = Mt3dms.load(namfile, model_ws=pth, verbose=True, exe_name="mt3dms")
    mt.model_ws = cpth
    ftlfile = "zeroth.ftl"
    mt.ftlfilename = ftlfile
    mt.write_input()
    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


@flaky(max_runs=3)
@requires_exe("mfnwt", "mt3dms")
@excludes_platform(
    "Windows", ci_only=True
)  # TODO remove once fixed in MT3D-USGS
def test_mfnwt_CrnkNic(function_tmpdir, mfnwtmt3d_model_path):
    pth = mfnwtmt3d_model_path / "sft_crnkNic"
    namefile = "CrnkNic.nam"
    mf = Modflow.load(
        namefile,
        model_ws=pth,
        version="mfnwt",
        verbose=True,
        exe_name="mfnwt",
    )

    cpth = function_tmpdir / "SFT_CRNKNIC"
    mf.model_ws = cpth

    mf.write_input()
    success, buff = mf.run_model(silent=False)
    assert success, f"{mf.name} did not run"

    namefile = "CrnkNic.mtnam"
    mt = Mt3dms.load(
        namefile,
        model_ws=pth,
        verbose=True,
        version="mt3d-usgs",
        exe_name="mt3dusgs",
    )

    mt.model_ws = cpth
    ftlfile = "CrnkNic.ftl"
    mt.ftlfilename = ftlfile
    mt.ftlfree = True
    mt.write_input()
    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


@pytest.mark.slow
@requires_exe("mfnwt", "mt3dusgs")
def test_mfnwt_LKT(function_tmpdir, mfnwtmt3d_model_path):
    pth = mfnwtmt3d_model_path / "lkt"
    namefile = "lkt_mf.nam"
    mf = Modflow.load(
        namefile,
        model_ws=pth,
        version="mfnwt",
        verbose=True,
        forgive=False,
        exe_name="mfnwt",
    )

    assert not mf.load_fail, "MODFLOW model did not load"

    cpth = function_tmpdir / "LKT"
    mf.model_ws = cpth

    # write modflow-nwt files
    mf.write_input()

    success = False
    success, buff = mf.run_model(silent=False)
    assert success, f"{mf.name} did not run"

    namefile = "lkt_mt.nam"
    mt = Mt3dms.load(
        namefile,
        model_ws=pth,
        verbose=True,
        version="mt3d-usgs",
        exe_name="mt3dusgs",
        modflowmodel=mf,
    )
    mt.model_ws = cpth
    ftlfile = "lkt.ftl"
    mt.ftlfilename = ftlfile
    mt.ftlfree = True

    # write mt3d files
    mt.write_input()

    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


@pytest.mark.slow
@requires_exe("mfnwt", "mt3dusgs")
def test_mfnwt_keat_uzf(function_tmpdir, mfnwtmt3d_model_path):
    pth = mfnwtmt3d_model_path / "keat_uzf"
    namefile = "Keat_UZF_mf.nam"
    mf = Modflow.load(
        namefile,
        model_ws=pth,
        version="mfnwt",
        verbose=True,
        exe_name="mfnwt",
    )

    cpth = function_tmpdir / "KEAT_UZF"
    mf.model_ws = cpth

    mf.write_input()
    success, buff = mf.run_model(silent=True)
    assert success, f"{mf.name} did not run"

    namefile = "Keat_UZF_mt.nam"
    mt = Mt3dms.load(
        namefile,
        model_ws=pth,
        verbose=True,
        version="mt3d-usgs",
        exe_name="mt3dusgs",
    )
    # Check a few options specified as optional keywords on line 3
    assert mt.btn.DRYCell is True
    assert mt.btn.Legacy99Stor is False
    assert mt.btn.MFStyleArr is True
    assert mt.btn.AltWTSorb is False

    mt.model_ws = cpth
    ftlfile = "Keat_UZF.ftl"
    mt.ftlfilename = ftlfile
    mt.ftlfree = True
    mt.write_input()
    success, buff = mt.run_model(silent=False, normal_msg="program completed.")
    assert success, f"{mt.name} did not run"


def test_mt3d_create_withmfmodel(function_tmpdir):
    model_ws = function_tmpdir

    # Create a MODFLOW model
    mf = Modflow(model_ws=model_ws)
    dis = ModflowDis(mf)
    lpf = ModflowLpf(mf)

    # Create MT3D model
    mt = Mt3dms(modflowmodel=mf, model_ws=model_ws)
    btn = Mt3dBtn(mt)
    adv = Mt3dAdv(mt)
    dsp = Mt3dDsp(mt)
    ssm = Mt3dSsm(mt)
    gcg = Mt3dRct(mt)
    rct = Mt3dGcg(mt)
    tob = Mt3dTob(mt)

    # Write the output
    mt.write_input()

    # confirm that MT3D files exist
    assert os.path.isfile(model_ws / f"{mt.name}.{btn.extension[0]}")
    assert os.path.isfile(model_ws / f"{mt.name}.{adv.extension[0]}")
    assert os.path.isfile(model_ws / f"{mt.name}.{dsp.extension[0]}")
    assert os.path.isfile(model_ws / f"{mt.name}.{ssm.extension[0]}")
    assert os.path.isfile(model_ws / f"{mt.name}.{gcg.extension[0]}")
    assert os.path.isfile(model_ws / f"{mt.name}.{rct.extension[0]}")
    assert os.path.isfile(model_ws / f"{mt.name}.{tob.extension[0]}")


def test_mt3d_create_woutmfmodel(function_tmpdir):
    model_ws = function_tmpdir

    # Create MT3D model
    mt = Mt3dms(model_ws=model_ws)
    btn = Mt3dBtn(
        mt,
        nlay=1,
        nrow=2,
        ncol=2,
        nper=1,
        delr=1.0,
        delc=1.0,
        htop=1.0,
        dz=1.0,
        laycon=0,
        perlen=1.0,
        nstp=1,
        tsmult=1.0,
    )
    adv = Mt3dAdv(mt)
    dsp = Mt3dDsp(mt)

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")

        ssm = Mt3dSsm(mt)

        wrn_msg = "mxss is None and modflowmodel is None."
        if len(w) > 0:
            print(f"Number of warnings: {len(w)}")
            ipos = -1
            for idx, wm in enumerate(w):
                print(wm.message)
                if wrn_msg in str(wm.message):
                    ipos = idx
                    break

        assert ipos >= 0, f"'{wrn_msg}' warning message not issued"
        assert (
            w[ipos].category == UserWarning
        ), f"Warning category: {w[0].category}"

    gcg = Mt3dRct(mt)
    rct = Mt3dGcg(mt)
    tob = Mt3dTob(mt)

    # Write the output
    mt.write_input()

    # confirm that MT3D files exist
    assert os.path.isfile(
        os.path.join(model_ws, f"{mt.name}.{btn.extension[0]}")
    )
    assert os.path.isfile(
        os.path.join(model_ws, f"{mt.name}.{adv.extension[0]}")
    )
    assert os.path.isfile(
        os.path.join(model_ws, f"{mt.name}.{dsp.extension[0]}")
    )
    assert os.path.isfile(
        os.path.join(model_ws, f"{mt.name}.{ssm.extension[0]}")
    )
    assert os.path.isfile(
        os.path.join(model_ws, f"{mt.name}.{gcg.extension[0]}")
    )
    assert os.path.isfile(
        os.path.join(model_ws, f"{mt.name}.{rct.extension[0]}")
    )
    assert os.path.isfile(
        os.path.join(model_ws, f"{mt.name}.{tob.extension[0]}")
    )


def test_mt3d_pht3d(function_tmpdir):
    # Note: this test is incomplete!
    model_ws = function_tmpdir

    # Create MT3D model
    mt = Mt3dms(model_ws=model_ws)
    phc = Mt3dPhc(mt, minkin=[[[1]], [[2]]])

    # Write the output
    mt.write_input()

    # confirm that MT3D files exist
    assert os.path.isfile(
        os.path.join(model_ws, f"{mt.name}.{phc.extension[0]}")
    )


def test_mt3d_multispecies(function_tmpdir):
    # modflow model
    modelname = "multispecies"
    nlay = 1
    nrow = 20
    ncol = 20
    nper = 10
    mf = Modflow(
        modelname=modelname,
        model_ws=function_tmpdir,
    )
    dis = ModflowDis(mf, nlay=nlay, nrow=nrow, ncol=ncol, nper=nper)
    lpf = ModflowLpf(mf)
    rch = ModflowRch(mf)
    evt = ModflowEvt(mf)
    mf.write_input()

    # Create a 5-component mt3d model and write the files
    ncomp = 5
    mt = Mt3dms(
        modelname=modelname,
        modflowmodel=mf,
        model_ws=function_tmpdir,
        verbose=True,
    )
    sconc3 = np.random.random((nrow, ncol))
    btn = Mt3dBtn(
        mt, ncomp=ncomp, sconc=1.0, sconc2=2.0, sconc3=sconc3, sconc5=5.0
    )
    # check obs I/O
    mt.btn.obs = np.array([[0, 2, 300], [0, 1, 250]])
    crch32 = np.random.random((nrow, ncol))
    cevt33 = np.random.random((nrow, ncol))
    ssm = Mt3dSsm(
        mt,
        crch=1.0,
        crch2=2.0,
        crch3={2: crch32},
        crch5=5.0,
        cevt=1.0,
        cevt2=2.0,
        cevt3={3: cevt33},
        cevt5=5.0,
    )
    crch2 = ssm.crch[1].array
    assert crch2.max() == 2.0
    cevt2 = ssm.cevt[1].array
    assert cevt2.max() == 2.0
    mt.write_input()

    # Create a second MODFLOW model
    modelname2 = "multispecies2"
    mf2 = Modflow(
        modelname=modelname2,
        model_ws=function_tmpdir,
    )
    dis2 = ModflowDis(mf2, nlay=nlay, nrow=nrow, ncol=ncol, nper=nper)

    # Load the MT3D model into mt2 and then write it out
    fname = f"{modelname}.nam"
    mt2 = Mt3dms.load(
        fname,
        model_ws=function_tmpdir,
        verbose=True,
    )
    # check obs I/O
    assert np.all(mt.btn.obs == mt2.btn.obs)
    mt2.name = modelname2
    mt2.write_input()


@requires_exe("mfnwt", "mt3dusgs")
def test_lkt_with_multispecies(function_tmpdir):
    # Bug discovered in LKT with multi-species.  Adding test to check this functionality

    modelname = "lkttest"
    mf = Modflow(
        modelname=modelname,
        exe_name="mfnwt",
        model_ws=function_tmpdir,
        version="mfnwt",
    )

    nrow = 44
    ncol = 55
    delr = 27500.0 / ncol
    delc = 22000.0 / nrow
    xmax = ncol * delr
    ymax = nrow * delc
    X, Y = np.meshgrid(
        np.linspace(delr / 2, xmax - delr / 2, ncol),
        np.linspace(ymax - delc / 2, 0 + delc / 2, nrow),
    )

    # output control (oc) package for MODFLOW-NWT
    oc = ModflowOc(mf)

    # solver package (Newton-Raphson Solver)
    nwt = ModflowNwt(
        mf,
        headtol=1.0e-4,
        fluxtol=5,
        maxiterout=5000,
        thickfact=1e-06,
        linmeth=2,
        iprnwt=1,
        ibotav=1,
        options="SIMPLE",
    )

    ## discretization (DIS) package
    top1 = np.array(
        [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            2.015692e02,
            2.01e02,
            2.02e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.95e02,
            1.96e02,
            1.96e02,
            1.97e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.015957e02,
            2.019755e02,
            2.02e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            3.52e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.935120e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.96e02,
            1.97e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.01e02,
            2.02e02,
            2.02e02,
            2.03e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            3.48e02,
            3.48e02,
            3.50e02,
            3.52e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.933555e02,
            1.94e02,
            1.95e02,
            1.95e02,
            1.96e02,
            1.97e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.019790e02,
            2.02e02,
            2.03e02,
            2.03e02,
            2.04e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            3.37e02,
            3.43e02,
            3.48e02,
            3.48e02,
            3.49e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.932550e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.97e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.06e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            3.36e02,
            3.36e02,
            3.43e02,
            3.48e02,
            3.48e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.933208e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.97e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.09e02,
            2.09e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            3.19e02,
            3.19e02,
            3.29e02,
            3.36e02,
            3.36e02,
            3.36e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.97e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            2.50e02,
            2.51e02,
            0,
            0,
            0,
            0,
            0,
            0,
            2.91e02,
            3.17e02,
            3.17e02,
            3.17e02,
            3.29e02,
            3.29e02,
            3.29e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.95e02,
            1.96e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.17e02,
            2.18e02,
            2.21e02,
            2.23e02,
            2.24e02,
            2.25e02,
            0,
            0,
            0,
            0,
            0,
            0,
            2.48e02,
            2.49e02,
            2.50e02,
            2.51e02,
            2.52e02,
            2.59e02,
            2.63e02,
            2.67e02,
            2.73e02,
            2.88e02,
            2.91e02,
            3.03e02,
            3.10e02,
            3.10e02,
            3.17e02,
            3.17e02,
            3.17e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.95e02,
            1.96e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.12e02,
            2.15e02,
            2.17e02,
            2.18e02,
            2.21e02,
            2.23e02,
            2.25e02,
            2.25e02,
            2.28e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.39e02,
            2.46e02,
            2.48e02,
            2.48e02,
            2.50e02,
            2.52e02,
            2.54e02,
            2.59e02,
            2.63e02,
            2.67e02,
            2.73e02,
            2.79e02,
            2.88e02,
            2.96e02,
            3.03e02,
            3.03e02,
            3.10e02,
            3.10e02,
            3.11e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.95e02,
            1.96e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.13e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.21e02,
            2.23e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.39e02,
            2.43e02,
            2.46e02,
            2.48e02,
            2.50e02,
            2.51e02,
            2.54e02,
            2.59e02,
            2.63e02,
            2.67e02,
            2.73e02,
            2.79e02,
            2.79e02,
            2.91e02,
            2.96e02,
            2.96e02,
            3.03e02,
            3.10e02,
            3.10e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.95e02,
            1.97e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.03e02,
            2.04e02,
            2.06e02,
            2.07e02,
            2.09e02,
            2.10e02,
            2.12e02,
            2.13e02,
            2.16e02,
            2.18e02,
            2.18e02,
            2.21e02,
            2.23e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.48e02,
            2.50e02,
            2.51e02,
            2.54e02,
            2.59e02,
            2.63e02,
            2.67e02,
            2.72e02,
            2.75e02,
            2.79e02,
            2.80e02,
            2.91e02,
            2.91e02,
            2.96e02,
            2.96e02,
            2.96e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.95e02,
            1.97e02,
            1.98e02,
            2.00e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.09e02,
            2.10e02,
            2.12e02,
            2.14e02,
            2.16e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.23e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.46e02,
            2.50e02,
            2.50e02,
            2.53e02,
            2.59e02,
            2.59e02,
            2.63e02,
            2.67e02,
            2.73e02,
            2.79e02,
            2.79e02,
            2.80e02,
            2.88e02,
            2.91e02,
            2.91e02,
            2.91e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.94e02,
            1.95e02,
            1.97e02,
            1.99e02,
            2.00e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.08e02,
            2.09e02,
            2.11e02,
            2.12e02,
            2.15e02,
            2.16e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.23e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.37e02,
            2.41e02,
            2.43e02,
            2.46e02,
            2.48e02,
            2.50e02,
            2.52e02,
            2.54e02,
            2.59e02,
            2.63e02,
            2.67e02,
            2.72e02,
            2.75e02,
            2.79e02,
            2.79e02,
            2.79e02,
            2.81e02,
            2.88e02,
            2.88e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.935360e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.98e02,
            1.99e02,
            2.01e02,
            2.02e02,
            2.04e02,
            2.05e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.11e02,
            2.12e02,
            2.15e02,
            2.16e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.23e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.31e02,
            2.32e02,
            2.35e02,
            2.36e02,
            2.39e02,
            2.43e02,
            2.43e02,
            2.48e02,
            2.50e02,
            2.50e02,
            2.53e02,
            2.59e02,
            2.63e02,
            2.63e02,
            2.67e02,
            2.72e02,
            2.75e02,
            2.75e02,
            2.79e02,
            2.79e02,
            2.79e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.937070e02,
            1.94e02,
            1.95e02,
            1.97e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.04e02,
            2.05e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.23e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.31e02,
            2.31e02,
            2.35e02,
            2.35e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.46e02,
            2.48e02,
            2.50e02,
            2.52e02,
            2.54e02,
            2.59e02,
            2.63e02,
            2.67e02,
            2.67e02,
            2.72e02,
            2.73e02,
            2.75e02,
            2.75e02,
            2.75e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.938565e02,
            1.94e02,
            1.96e02,
            1.97e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.03e02,
            2.04e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.23e02,
            2.25e02,
            2.25e02,
            2.28e02,
            2.30e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.46e02,
            2.48e02,
            2.50e02,
            2.50e02,
            2.53e02,
            2.59e02,
            2.59e02,
            2.63e02,
            2.67e02,
            2.67e02,
            2.68e02,
            2.72e02,
            2.73e02,
            2.73e02,
            2.73e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.94e02,
            1.96e02,
            1.97e02,
            1.99e02,
            2.00e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.18e02,
            2.19e02,
            2.21e02,
            2.23e02,
            2.24e02,
            2.25e02,
            2.28e02,
            2.30e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.36e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.46e02,
            2.48e02,
            2.50e02,
            2.51e02,
            2.53e02,
            2.59e02,
            2.59e02,
            2.63e02,
            2.67e02,
            2.67e02,
            2.67e02,
            2.69e02,
            2.72e02,
            2.72e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.06e02,
            2.07e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.18e02,
            2.19e02,
            2.21e02,
            2.22e02,
            2.24e02,
            2.25e02,
            2.28e02,
            2.28e02,
            2.31e02,
            2.32e02,
            2.35e02,
            2.35e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.46e02,
            2.48e02,
            2.50e02,
            2.50e02,
            2.52e02,
            2.54e02,
            2.59e02,
            2.59e02,
            2.63e02,
            2.63e02,
            2.67e02,
            2.67e02,
            0,
            0,
            0,
            0,
            1.930262e02,
            1.930813e02,
            1.931463e02,
            1.932345e02,
            1.933725e02,
            1.938515e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.97e02,
            1.98e02,
            1.99e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.18e02,
            2.18e02,
            2.21e02,
            2.22e02,
            2.24e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.31e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.36e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.46e02,
            2.48e02,
            2.50e02,
            2.50e02,
            2.52e02,
            2.54e02,
            2.59e02,
            2.59e02,
            2.63e02,
            2.63e02,
            0,
            0,
            0,
            0,
            0,
            1.930745e02,
            1.932300e02,
            1.934108e02,
            1.936452e02,
            1.939805e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.97e02,
            1.98e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.14e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.20e02,
            2.22e02,
            2.23e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.30e02,
            2.31e02,
            2.32e02,
            2.35e02,
            2.35e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.43e02,
            2.46e02,
            2.48e02,
            2.50e02,
            2.50e02,
            2.51e02,
            2.54e02,
            2.59e02,
            2.59e02,
            0,
            0,
            0,
            0,
            0,
            0,
            1.931133e02,
            1.933483e02,
            1.936155e02,
            1.939447e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.97e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.14e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.23e02,
            2.24e02,
            2.25e02,
            2.28e02,
            2.28e02,
            2.31e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.36e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.43e02,
            2.46e02,
            2.48e02,
            2.50e02,
            2.50e02,
            2.50e02,
            2.58e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.931405e02,
            1.934308e02,
            1.937560e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.95e02,
            1.96e02,
            1.97e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.17e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.22e02,
            2.24e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.30e02,
            2.31e02,
            2.32e02,
            2.35e02,
            2.35e02,
            2.36e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.43e02,
            2.46e02,
            2.46e02,
            2.48e02,
            2.48e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.931537e02,
            1.934730e02,
            1.938313e02,
            1.94e02,
            1.947535e02,
            1.95e02,
            1.96e02,
            1.96e02,
            1.97e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.18e02,
            2.18e02,
            2.20e02,
            2.22e02,
            2.23e02,
            2.25e02,
            2.25e02,
            2.28e02,
            2.28e02,
            2.31e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.35e02,
            2.36e02,
            2.39e02,
            2.41e02,
            2.43e02,
            2.43e02,
            2.43e02,
            2.46e02,
            2.46e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.931505e02,
            1.934700e02,
            1.938398e02,
            1.94e02,
            1.947937e02,
            1.95e02,
            1.96e02,
            1.968170e02,
            1.97e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.23e02,
            2.24e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.30e02,
            2.31e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.35e02,
            2.36e02,
            2.39e02,
            2.41e02,
            2.41e02,
            2.43e02,
            2.43e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.931243e02,
            1.934108e02,
            1.937765e02,
            1.94e02,
            1.947485e02,
            1.95e02,
            1.96e02,
            1.967558e02,
            1.97e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.14e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.22e02,
            2.23e02,
            2.25e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.30e02,
            2.31e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.35e02,
            2.36e02,
            2.39e02,
            2.41e02,
            2.41e02,
            2.41e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.932635e02,
            1.936290e02,
            1.94e02,
            1.946230e02,
            1.95e02,
            1.96e02,
            1.96e02,
            1.97e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.18e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.23e02,
            2.24e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.28e02,
            2.30e02,
            2.31e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.35e02,
            2.36e02,
            2.39e02,
            2.39e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.933783e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.957018e02,
            1.96e02,
            1.97e02,
            1.98e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.22e02,
            2.23e02,
            2.24e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.28e02,
            2.30e02,
            2.31e02,
            2.31e02,
            2.34e02,
            2.35e02,
            2.35e02,
            2.35e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.935405e02,
            1.94e02,
            1.95e02,
            1.95e02,
            1.96e02,
            1.97e02,
            1.97e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.14e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.23e02,
            2.24e02,
            2.25e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.28e02,
            2.30e02,
            2.31e02,
            2.31e02,
            2.32e02,
            2.34e02,
            2.34e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.932890e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.96e02,
            1.97e02,
            1.98e02,
            1.99e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.22e02,
            2.23e02,
            2.24e02,
            2.25e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.28e02,
            2.30e02,
            2.31e02,
            2.31e02,
            2.31e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.935508e02,
            1.94e02,
            1.95e02,
            1.95e02,
            1.96e02,
            1.97e02,
            1.97e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.22e02,
            2.23e02,
            2.24e02,
            2.25e02,
            2.25e02,
            2.27e02,
            2.28e02,
            2.28e02,
            2.28e02,
            2.30e02,
            2.30e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.932863e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.96e02,
            1.97e02,
            1.98e02,
            1.98e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.19e02,
            2.20e02,
            2.21e02,
            2.22e02,
            2.23e02,
            2.24e02,
            2.25e02,
            2.25e02,
            2.27e02,
            2.27e02,
            2.28e02,
            2.28e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.935100e02,
            1.94e02,
            1.95e02,
            1.95e02,
            1.96e02,
            1.96e02,
            1.97e02,
            1.98e02,
            1.98e02,
            1.99e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.04e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.21e02,
            2.22e02,
            2.23e02,
            2.24e02,
            2.25e02,
            2.25e02,
            2.25e02,
            2.27e02,
            2.27e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.932413e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.95e02,
            1.96e02,
            1.96e02,
            1.97e02,
            1.97e02,
            1.99e02,
            2.00e02,
            2.01e02,
            2.02e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.18e02,
            2.20e02,
            2.21e02,
            2.22e02,
            2.22e02,
            2.23e02,
            2.24e02,
            2.24e02,
            2.25e02,
            2.25e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.933477e02,
            1.94e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.95e02,
            1.96e02,
            1.96e02,
            1.97e02,
            1.99e02,
            2.00e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.15e02,
            2.16e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.19e02,
            2.20e02,
            2.21e02,
            2.22e02,
            2.22e02,
            2.23e02,
            2.24e02,
            2.24e02,
            2.24e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.932182e02,
            1.933268e02,
            1.94e02,
            1.94e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.98e02,
            1.99e02,
            2.01e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.13e02,
            2.14e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.18e02,
            2.19e02,
            2.20e02,
            2.21e02,
            2.21e02,
            2.22e02,
            2.23e02,
            2.23e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.93e02,
            1.93e02,
            1.93e02,
            1.93e02,
            1.93e02,
            1.94e02,
            1.94e02,
            1.95e02,
            1.97e02,
            1.99e02,
            2.00e02,
            2.02e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.14e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.18e02,
            2.19e02,
            2.20e02,
            2.21e02,
            2.21e02,
            2.22e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.94e02,
            1.96e02,
            1.98e02,
            2.00e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.14e02,
            2.15e02,
            2.16e02,
            2.16e02,
            2.17e02,
            2.18e02,
            2.18e02,
            2.19e02,
            2.20e02,
            2.21e02,
            2.21e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.95e02,
            1.97e02,
            1.99e02,
            2.01e02,
            2.03e02,
            2.04e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.14e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.17e02,
            2.18e02,
            2.18e02,
            2.19e02,
            2.20e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.95e02,
            1.97e02,
            1.99e02,
            2.01e02,
            2.03e02,
            2.04e02,
            2.05e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.14e02,
            2.15e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.17e02,
            2.18e02,
            2.18e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.936795e02,
            1.94e02,
            1.95e02,
            1.97e02,
            1.99e02,
            2.01e02,
            2.02e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.12e02,
            2.13e02,
            2.14e02,
            2.15e02,
            2.15e02,
            2.16e02,
            2.17e02,
            2.17e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.938033e02,
            1.94e02,
            1.96e02,
            1.97e02,
            1.99e02,
            2.01e02,
            2.02e02,
            2.04e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.14e02,
            2.14e02,
            2.15e02,
            2.16e02,
            2.16e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.938665e02,
            1.94e02,
            1.96e02,
            1.98e02,
            1.99e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.13e02,
            2.14e02,
            2.14e02,
            2.15e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.94e02,
            1.94e02,
            1.96e02,
            1.98e02,
            1.99e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.11e02,
            2.11e02,
            2.12e02,
            2.13e02,
            2.13e02,
            2.14e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            1.933480e02,
            1.94e02,
            1.95e02,
            1.96e02,
            1.98e02,
            1.99e02,
            2.01e02,
            2.02e02,
            2.03e02,
            2.05e02,
            2.06e02,
            2.07e02,
            2.08e02,
            2.09e02,
            2.10e02,
            2.107930e02,
            2.11e02,
            2.12e02,
            2.13e02,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
        ]
    )

    top1 = top1.reshape((nrow, ncol))
    bot1 = np.ones(top1.shape) * 180.0
    bot2 = np.ones(top1.shape) * 160.0
    bot3 = np.ones(top1.shape) * 140.0
    botm = [bot1, bot2, bot3]
    botm = np.array(botm)

    # Create the discretization object
    # itmuni = 4 (days); lenuni = 2 (meters)
    dis = ModflowDis(
        mf,
        3,
        nrow,
        ncol,
        nper=1,
        delr=delr,
        delc=delc,
        top=top1,
        botm=botm,
        laycbd=0,
        itmuni=4,
        lenuni=2,
        steady=True,
        nstp=1,
        tsmult=1.0,
        perlen=86400.0,
    )

    ## Instantiate upstream weighting (UPW) flow package
    upw = ModflowUpw(
        mf,
        laytyp=[1, 1, 1],  # >0: convertible
        layavg=0,  #  0: harmonic mean
        chani=1.0,  # >0: CHANI is the horizontal anis for entire layer
        layvka=0,  # =0: indicates VKA is vertical hydraulic conductivity
        laywet=0,  # Always set equal to zero in UPW package
        ipakcb=53,
        hdry=-1.0e30,
        iphdry=0,
        hk=3.172e-03,
        vka=3.172e-04,  # Is equal to vert. K b/c LAYVKA = 0
        ss=0.00001,
        sy=0.30,
    )

    ## Instantiate basic (BAS or BA6) package for MODFLOW-NWT

    ibnd1 = [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ]

    ibnd2 = [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ]

    ibnd3 = [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ]

    sthd1 = [
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        193.0,
        193.0,
        193.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        193.0,
        193.0,
        193.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        193.0,
        193.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
        210.0,
    ]

    sthd1 = np.array(sthd1).reshape(top1.shape)
    sthd2 = np.ones(top1.shape) * 210.0
    sthd3 = sthd2.copy()

    bas = ModflowBas(
        mf,
        ibound=np.array(
            [
                np.array(ibnd1).reshape(top1.shape),
                np.array(ibnd2).reshape(top1.shape),
                np.array(ibnd3).reshape(top1.shape),
            ]
        ),
        hnoflo=-9999,
        strt=np.array([sthd1, sthd2, sthd3]),
    )

    ## Instantiate Lake (LAK) package for MODFLOW-NWT

    lkarr1 = [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        2,
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        3,
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        3,
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        3,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ]

    lkarr2 = [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        2,
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        2,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
    ]

    lkarr1 = np.array(lkarr1)
    lkarr2 = np.array(lkarr2)
    lkarr1 = lkarr1.reshape(top1.shape)
    lkarr2 = lkarr2.reshape(top1.shape)
    lkarr3 = np.zeros(top1.shape)
    lkarr = np.array([lkarr1, lkarr2, lkarr3])

    bdlk1 = np.zeros(top1.shape)
    bdlk2 = np.zeros(top1.shape)
    bdlk3 = np.zeros(top1.shape)
    bdlk1[bdlk1 == 1] = 1e-8
    bdlk1[bdlk1 == 2] = 2e-8
    bdlk1[bdlk1 == 3] = 1.5e-8
    bdlk2[bdlk2 == 1] = 1e-8
    bdlk2[bdlk2 == 2] = 2e-8
    bdlk2[bdlk2 == 3] = 1.5e-8

    lak = ModflowLak(
        mf,
        nlakes=int(np.max(lkarr)),
        ipakcb=3,  # From above
        theta=-1.0,  # Implicit
        nssitr=99,  # Maximum number of iterations for Newtons method
        sscncr=1.000e-02,  # Convergence criterion for equilibrium lake stage solution
        surfdep=2.000e-01,  # Height of small topological variations in lake-bottom
        stages=[200.0, 200.0, 200.0],
        stage_range=[
            (160.0, 220.0),
            (170.0, 220.0),
            (180.0, 220.0),
        ],  # Initial stage of each lake at the beginning of the run
        lakarr=lkarr,
        bdlknc=np.array([bdlk1, bdlk2, bdlk3]),
        flux_data={
            0: [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ]
        },
        unit_number=16,
    )

    # linkage with mass transport routing (LMT) package
    lmt = ModflowLmt(
        mf,
        output_file_name="lkttest.ftl",
        output_file_header="extended",
        output_file_format="formatted",
        package_flows=["lak"],
    )

    ## Now work on MT3D-USGS file creation

    mt = Mt3dms(
        modflowmodel=mf,
        modelname=modelname,
        model_ws=function_tmpdir,
        version="mt3d-usgs",
        namefile_ext="mtnam",
        exe_name="mt3dusgs",
        ftlfilename="lkttest.ftl",
        ftlfree=True,
    )

    ## Instantiate basic transport (BTN) package for MT3D-USGS

    ncomp = 2
    mcomp = 2
    lunit = "FT"
    laycon = 1
    sconc = 0.0
    sconc2 = 0.0
    prsity = 0.3
    cinact = -1.0
    thkmin = 0.01
    nprs = -1
    nprobs = 1
    nprmas = 1
    dt0 = 0.0
    nstp = 1
    mxstrn = 50000
    tsmult = 1
    ttsmult = 1.0
    ttsmax = 0
    perlen = 86400.0

    btn = Mt3dBtn(
        mt,
        lunit=lunit,
        ncomp=ncomp,
        mcomp=mcomp,
        sconc=sconc,
        prsity=prsity,
        cinact=cinact,
        laycon=laycon,
        thkmin=thkmin,
        nprs=nprs,
        nprobs=nprobs,
        chkmas=True,
        nprmas=nprmas,
        perlen=perlen,
        dt0=dt0,
        nstp=nstp,
        tsmult=tsmult,
        mxstrn=mxstrn,
        ttsmult=ttsmult,
        ttsmax=ttsmax,
        sconc2=sconc2,
    )

    ## Instantiate advection (ADV) package for MT3D-USGS

    mixelm = 0
    percel = 0.7500
    mxpart = 5000
    nadvfd = 1  # (1 = Upstream weighting)

    adv = Mt3dAdv(
        mt, mixelm=mixelm, percel=percel, mxpart=mxpart, nadvfd=nadvfd
    )

    ## Instantiate generalized conjugate gradient solver (GCG)
    # package for MT3D-USGS

    mxiter = 1
    iter1 = 50
    isolve = 1
    ncrs = 0
    accl = 1.000000
    cclose = 1.00e-05
    iprgcg = 0

    gcg = Mt3dGcg(
        mt,
        mxiter=mxiter,
        iter1=iter1,
        isolve=isolve,
        ncrs=ncrs,
        accl=accl,
        cclose=cclose,
        iprgcg=iprgcg,
    )

    ## Instantiate source-sink mixing (SSM) package for MT3D-USGS

    mxss = 0
    ssm = Mt3dSsm(mt, mxss=mxss)

    ## Instantiate LKT package

    nlkinit = 3
    mxlkbc = 12
    icbclk = 0
    ietlak = 0
    coldlak = [2.0, 1.0, 6.0]
    coldlak2 = [3.0, 2.0, 7.0]  # Starting concentration for species 2

    lkt_flux_data = {0: [[0, 1, 0.0, 0.0], [1, 1, 0.0, 0.0], [2, 1, 0.0, 0.0]]}

    lkt = Mt3dLkt(
        mt,
        nlkinit=nlkinit,
        mxlkbc=mxlkbc,
        icbclk=icbclk,
        ietlak=ietlak,
        coldlak=coldlak,
        lk_stress_period_data=lkt_flux_data,
        coldlak2=coldlak2,
    )

    mf.write_input()
    mt.write_input()

    # Make sure the just written files are loadable
    namfile = f"{modelname}.nam"
    mf = Modflow.load(
        namfile,
        model_ws=function_tmpdir,
        version="mfnwt",
        verbose=True,
        exe_name="mfnwt",
    )
    namfile = f"{modelname}.mtnam"
    mt = Mt3dms.load(
        namfile,
        model_ws=function_tmpdir,
        verbose=True,
        version="mt3d-usgs",
        exe_name="mt3dusgs",
    )


def test_mt3dms_load_when_nam_dne():
    with pytest.raises(OSError):
        Mt3dms.load("nonexistent.nam")


@requires_exe("mf2005")
def test_mt3d_ssm_with_nodata_in_1st_sp(function_tmpdir):
    nlay, nrow, ncol = 3, 5, 5
    perlen = np.zeros((10), dtype=float) + 10
    nper = len(perlen)

    ibound = np.ones((nlay, nrow, ncol), dtype=int)

    botm = np.arange(-1, -4, -1)
    top = 0.0

    # creating MODFLOW model
    modelname = "model_mf"
    mf = Modflow(modelname, model_ws=function_tmpdir, exe_name="mf2005")
    dis = ModflowDis(
        mf,
        nlay=nlay,
        nrow=nrow,
        ncol=ncol,
        perlen=perlen,
        nper=nper,
        botm=botm,
        top=top,
        steady=False,
    )

    bas = ModflowBas(mf, ibound=ibound, strt=top)
    lpf = ModflowLpf(mf, hk=100, vka=100, ss=0.00001, sy=0.1)
    oc = ModflowOc(mf)
    pcg = ModflowPcg(mf)
    lmt = ModflowLmt(mf)

    # recharge
    rchrate = {}
    rchrate[0] = 0.0
    rchrate[5] = 0.001
    rchrate[6] = 0.0

    rch = ModflowRch(mf, rech=rchrate, nrchop=3)

    # define itype
    itype = Mt3dSsm.itype_dict()
    ssm_data = {}

    # Now for the point of this test: Enter SSM data sometime
    # after the first stress period (this was crashing flopy
    # version 3.2.13
    ghb_data = {}
    ssm_data[2] = [(nlay - 1, 4, 4, 1.0, itype["GHB"], 1.0, 100.0)]
    ghb_data[2] = [(nlay - 1, 4, 4, 0.1, 1.5)]
    ghb_data[5] = [(nlay - 1, 4, 4, 0.25, 1.5)]
    ssm_data[5] = [(nlay - 1, 4, 4, 0.5, itype["GHB"], 0.5, 200.0)]

    for k in range(nlay):
        for i in range(nrow):
            ghb_data[2].append((k, i, 0, 0.0, 100.0))
            ssm_data[2].append((k, i, 0, 0.0, itype["GHB"], 0.0, 0.0))

    ghb_data[5] = [(nlay - 1, 4, 4, 0.25, 1.5)]
    ssm_data[5] = [(nlay - 1, 4, 4, 0.5, itype["GHB"], 0.5, 200.0)]
    for k in range(nlay):
        for i in range(nrow):
            ghb_data[5].append((k, i, 0, -0.5, 100.0))
            ssm_data[5].append((k, i, 0, 0.0, itype["GHB"], 0.0, 0.0))

    ghb = ModflowGhb(mf, stress_period_data=ghb_data)

    # create MT3D-USGS model
    modelname = "model_mt"
    mt = Mt3dms(
        modflowmodel=mf,
        modelname=modelname,
        model_ws=function_tmpdir,
        exe_name="mt3dms",
    )
    btn = Mt3dBtn(mt, sconc=0, ncomp=2, sconc2=50.0)
    adv = Mt3dAdv(mt)
    ssm = Mt3dSsm(mt, stress_period_data=ssm_data)
    gcg = Mt3dGcg(mt)

    # Write the output
    mf.write_input()
    mt.write_input()

    # run the models
    success, buff = mf.run_model(report=True)
    assert success, "MODFLOW did not run"
    success, buff = mt.run_model(report=True, normal_msg="Program completed.")
    assert success, "MT3D did not run"

    ws = function_tmpdir / "ws2"
    mf2 = Modflow.load(
        "model_mf.nam", model_ws=function_tmpdir, exe_name="mf2005"
    )
    mf2.change_model_ws(ws)
    mt2 = Mt3dms.load(
        "model_mt.nam",
        model_ws=function_tmpdir,
        verbose=True,
        exe_name="mt3dms",
    )
    mt2.change_model_ws(ws)
    mf2.write_input()
    mt2.write_input()
    success, buff = mf2.run_model(report=True)
    assert success, "MODFLOW did not run"
    success, buff = mt2.run_model(report=True, normal_msg="Program completed.")
    assert success, "MT3D did not run"

    conca = UcnFile(os.path.join(function_tmpdir, "MT3D001.UCN")).get_alldata()
    concb = UcnFile(os.path.join(ws, "MT3D001.UCN")).get_alldata()

    assert np.allclose(conca, concb)


@requires_exe("mf2005")
def test_none_spdtype(function_tmpdir):
    # ensure that -1 and None work as valid list entries in the
    # stress period dictionary

    mf = Modflow(model_ws=function_tmpdir, exe_name="mf2005")
    dis = ModflowDis(mf, nper=2)
    bas = ModflowBas(mf)
    lpf = ModflowLpf(mf)
    spd = {0: -1, 1: None}
    wel = ModflowWel(mf, stress_period_data=spd)
    pcg = ModflowPcg(mf)
    mf.write_input()
    mf2 = Modflow.load(
        "modflowtest.nam", model_ws=function_tmpdir, verbose=True
    )
    success, buff = mf.run_model(report=True)
    assert success


def test_ssm_readwrite(function_tmpdir, example_data_path):
    # Instantiate MODFLOW model
    mf = Modflow()

    # These dimensions work for the SSM file to be loaded
    dis = ModflowDis(
        mf, nlay=10, nrow=118, ncol=153, nper=100, delr=100, delc=100, perlen=1
    )
    bas = ModflowBas(mf)

    # Instantiate MT3D model
    mt = Mt3dms(modflowmodel=mf, modelname="tran_v1", version="mt3d-usgs")
    btn = Mt3dBtn(mt)

    # Point to ssm file for test load and write
    pth = example_data_path / "ssm_load_test"
    fl = pth / "tran_v1_b1.ssm"

    # Check that example input file with no data specified for crch works
    # (file comes from: https://github.com/modflowpy/flopy/issues/743)
    ssm = Mt3dSsm.load(fl, mt)

    mt.change_model_ws(function_tmpdir)

    # Ensure file is writeable
    ssm.write_file()
